#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include "gc.h"

#define BUFSIZE 1024
static char buf[BUFSIZE];

/* Length of array */
#define LENGTHOF(x) (*( ((long *) (x)) - 1 ))

#define ASSERT(x,msg) if (!(x)) { printf("*** libic *** "); printf msg; exit(1); }

/**************************************************************
 * Internal routines
 */

char *__LIB_allocateArray(long len)
{
  ASSERT((len >= 0), ("__allocate: negative array length: %ld\n", len));

  long size = (len + 1) * sizeof(long);
  char *p = (char*) GC_malloc (size);
  ASSERT((p != NULL), ("GC_malloc: memory allocation failed\n"));
  memset (p, 0, size); 

  *((long*)p) = len;
  return (p + sizeof(long));
}

char *__LIB_allocateString(long len)
{
  ASSERT((len >= 0), ("__allocate: negative string length: %ld\n", len));

  long size = len + sizeof(long);
  char *p = (char*) GC_malloc (size);
  ASSERT((p != NULL), ("GC_malloc: memory allocation failed\n"));
  memset (p, 0, size); 

  *((long*)p) = len;
  return (p + sizeof(long));
}

char *__LIB_allocateObject(long size)
{
  ASSERT((size >= 0), ("__allocate: negative size: %ld\n", size));

  char *p = (char*) GC_malloc (size);
  ASSERT((p != NULL), ("GC_malloc: memory allocation failed\n"));
  memset (p, 0, size);

  return p;
}

char *__LIB_stringCat(char *s, char *d)
{
  long lens, lend, len;
  char *r;

  ASSERT((s != NULL), ("__stringcat: first source string is null\n"));
  ASSERT((d != NULL), ("__stringcat: second source string is null\n"));

  lens = LENGTHOF(s);
  lend = LENGTHOF(d);
  len = lens + lend;
  r = __LIB_allocateString(len);
  memcpy(r, s, lens);
  memcpy(r + lens, d, lend);
  return r;
}

char *__LIB_makeString(char *s)
{
  long len;
  char *d;

  ASSERT(s != NULL, ("__makestring: source string is null\n"));

  len = strlen(s);
  d = __LIB_allocateString(len);
  memcpy(d, s, len);
  return d;
}

/**************************************************************
 * io module
 */

void __LIB_print(char *s) 
{ 
  ASSERT(s != NULL, ("print: source string is null\n"));
  fwrite (s, sizeof(char), LENGTHOF(s), stdout); 
  fflush(stdout); 
}

void __LIB_println(char *s) 
{ 
  ASSERT(s != NULL, ("print: source string is null\n"));
  fwrite (s, sizeof(char), LENGTHOF(s), stdout);
  fwrite ("\n", sizeof(char), 1, stdout);
  fflush(stdout); 
}

void __LIB_printi(long i)  {   fprintf(stdout,"%ld", i); }
void __LIB_printb(char b)  {   fprintf(stdout,"%s", b?"true":"false"); }

long __LIB_eof()   {   return feof(stdin) ? 1 : 0; } 
long __LIB_readi() {   return getc(stdin); }

char *__LIB_readln() 
{ 
  fgets(buf,BUFSIZE-1,stdin); 
  buf[strlen(buf)-1] = 0;
  return __LIB_makeString(buf);  
}

/***************************************************************
 * conv module
 */

long __LIB_stoi(char *string, long error)
{
  
  int len = LENGTHOF(string);
  char z[len+1];  // make null-terminated string
  memcpy(z,string,len);
  z[len]=0;
  long i = atoi (z);
  return (i == 0 && z[0] != '0') ? error : i;
}

char *__LIB_itos(long i)
{
  sprintf (buf, "%ld", i);
  return __LIB_makeString(buf);
}

char *__LIB_atos(long *a)
{
  long len, i;
  char *p;
  ASSERT(a != NULL, ("atos: null source array\n"));
  len = LENGTHOF(a);
  p = __LIB_allocateString(len);
  for(i=0; i<len; i++) p[i] = (char) a[i];
  return p;
}

long *__LIB_stoa(char *s)
{
  long len, i, *p;
  ASSERT(s != NULL, ("stoa: null source string\n"));
  len = LENGTHOF(s);
  p = (long*)__LIB_allocateArray(len);
  for(i=0; i<len; i++) p[i] = (long) s[i];
  return p;
}

/*************************************************************
 * random, exit, timing
 */

void __LIB_exit(long n) 
{ 
  exit(n); 
}

long __LIB_random(long n) 
{ 
  ASSERT((n > 0), ("random: argument is not positive: %ld\n", n));

  return (long) 
    ((double)n *
     ((double)rand()/(double)RAND_MAX)); 
}

long __LIB_time() 
{ 
  return (long) 
    (1000.0 * 
     ((double)clock())/(double)CLOCKS_PER_SEC);
}

/*************************************************************
 * main
 */

int main (int argc, char **argv)
{
  char **icarg;
  long i;
  extern long (_ic_main) (char**); 

  argc--;
  argv++;
  
  icarg = (char **) __LIB_allocateArray(argc);
  for (i = 0; i < argc; i++) 
    icarg[i] = __LIB_makeString(argv[i]);
  
  __ic_main(icarg); 
}
